{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/BS5/TEXConv/src/TEXConv.dpr $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2014-01-13 23:50:56 +0100 (Mo, 13 Jan 2014) $
  Revision: $Rev: 247 $

  Broken Sword 5 TEX converter.
}
program TEXConv;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  ModInfo in 'utils\ModInfo.pas',
  SaveTGA in 'utils\SaveTGA.pas',
  SysUtils,
  StrUtils,
  ZLib;

const
  Command_None    = $00000000;
  Command_Invalid = $00000001;
  Command_Info    = $00000002;
  Command_Operate = $00000004;

  TEXHeaderID = 'RTEX';

type
  TParamRec = record
    FileSpec   : AnsiString;
    Info       : Boolean;
  end;

  TTEXHeader = packed record
    ID        : array[0..3] of Char;
    ColorDepth: Longword;
    Width     : Longword;
    Height    : Longword;
    UncompSize: Longword;
    CompSize  : Longword;
  end;

  TBuffer = array of Byte;

var
  ParamRec   : TParamRec;
  Command    : Longword;

function OpenInFile(var InFile: file; InName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; OutName: string): Boolean;
var
  Dir: AnsiString;
begin
  // first create the directory portion of the output file
  Dir := ExtractFileDir(ExpandFileName(OutName));
  ForceDirectories(Dir);

  // open the output file
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function ReadHeader(var TEXFile: file; out TEXHeader: TTEXHeader): Boolean;
var
  BytesRead: Longint;
begin
  BlockRead(TEXFile, TEXHeader, SizeOf(TEXHeader), BytesRead);
  Result := ((BytesRead = SizeOf(TEXHeader)) and (TEXHeader.ID = TEXHeaderID))
end;

function Decompress(const InBuffer: TBuffer; InOfs, InLen: Longint;
  var OutBuffer: TBuffer; OutOfs, OutLen: Longint): Boolean;
var
  DecompSize: Longint;
  TempBuffer: Pointer;
begin
  // standard ZLib decompression
  DecompressBuf(@InBuffer[InOfs], InLen, OutLen, TempBuffer,
    DecompSize);

  Result := (DecompSize = OutLen);
  if Result then
    Move(TempBuffer^, OutBuffer[OutOfs], DecompSize);

  FreeMem(TempBuffer, DecompSize)
end;

procedure ConvertFile(var TEXFile: file; const TEXName: AnsiString;
  const TEXHeader: TTEXHeader);
var
  i        : Longint;
  OutName  : AnsiString;
  Buffer,
  UncompBuf: TBuffer;
  RefSize,
  Color    : Longword;
begin
  // determine a name for this file
  OutName := TEXName + '.tga';

  // do some sanity checks ...
  with TEXHeader do
    RefSize := Width * Height * ColorDepth;

  if TEXHeader.UncompSize <> RefSize then
    WriteLn(Format(' warning: processing %d bytes of image data, expected %d',
      [TEXHeader.UncompSize, RefSize]));

  if TEXHeader.ColorDepth <> 4 then
    WriteLn('  warning: possibly unsupported color depth');

  if TEXHeader.CompSize = 0 then
    // uncompressed file
    SetLength(Buffer, TEXHeader.UncompSize)
  else
    // compressed file
    SetLength(Buffer, TEXHeader.CompSize);

  // read file data
  {$I-}
  BlockRead(TEXFile, Buffer[0], Length(Buffer));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  read error');

    SetLength(Buffer, 0);
    Exit
  end;

  // optionally decompress a compressed file
  if TEXHeader.CompSize > 0 then
  begin
    SetLength(UncompBuf, TEXHeader.UncompSize);

    if Decompress(Buffer, 0, Length(Buffer), UncompBuf, 0,
      Length(UncompBuf)) then
      Buffer := UncompBuf
    else
      WriteLn('  decompression error')
  end;

  // convert color words from RGBA to BGRA
  for i := 0 to High(Buffer) shr 2 do
  begin
    Move(Buffer[i shl 2], Color, SizeOf(Color));
    Color :=    ((Color and $00ff0000) shr 16)
             or ((Color and $000000ff) shl 16)
             or (Color and $ff00ff00);
    Move(Color, Buffer[i shl 2], SizeOf(Color))
  end;

  // convert and save image
  if not DumpTGA(OutName, TEXHeader.Width, TEXHeader.Height, @Buffer[0],
    TGAPixelDepth_32Bit) then
    WriteLn('  error dumping TGA');

  SetLength(Buffer, 0);
  SetLength(UncompBuf, 0)
end;

procedure ScanFiles;
var
  FileName,
  Dir,
  TEXName  : string;
  TEXFile  : file;
  TEXHeader: TTEXHeader;
  SearchRec: TSearchRec;
begin
  // get the directory information
  Dir := ExtractFilePath(ParamRec.FileSpec);

  // search for all files matching the file specification
  if FindFirst(ParamRec.FileSpec, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      TEXName := Dir + FileName;

      WriteLn(TEXName);

      if OpenInFile(TEXFile, TEXName) then
      begin
        if ReadHeader(TEXFile, TEXHeader) then
          ConvertFile(TEXFile, FileName, TEXHeader)
        else
          WriteLn('  not a valid TEX file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [ParamRec.FileSpec]));

  SysUtils.FindClose(SearchRec)
end;

function ProcessParams: Longword;
var
  i         : Longint;
  Param,
  ParamLower: AnsiString;
begin
  Result := Command_None;

  if ParamCount = 0 then
  begin
    Result := Result or Command_Invalid;
    Exit
  end;

  with ParamRec do
  begin
    FileSpec    := '';
    Info        := False;

    for i := 1 to ParamCount do
    begin
      Param := ParamStr(i);
      ParamLower := AnsiLowerCase(Param);
      if ParamLower = '/info' then
        Info := True
      else if FileSpec = '' then
        FileSpec := Param
      else
        Result := Result or Command_Invalid
    end;

    if Info then
      Result := Result or Command_Info;
    if (FileSpec <> '') and ((Result and Command_Invalid) = 0) then
      Result := Result or Command_Operate
  end
end;

begin
  Command := ProcessParams;

  if (Command and Command_Info) <> 0 then
    WriteLn(GetModuleFileVersionString);
  if (Command and Command_Invalid) <> 0 then
    WriteLn('Usage: TEXConvd <TEXFiles> [/info]');
  if (Command and Command_Operate) <> 0 then
    ScanFiles
end.
