{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/BS5/TEXConv/src/utils/SaveTGA.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2014-01-13 23:21:19 +0100 (Mo, 13 Jan 2014) $
  Revision: $Rev: 242 $

  Library providing functions to export data as a TGA bitmap file.
}
unit SaveTGA;

{$DEFINE WRITE_TGA_FOOTER}
{$DEFINE CONVERT_WHEN_NECESSARY}

interface

type
  TTGACompression = (TGACompression_None, TGACompression_RLE);
  TTGAPixelDepth = (TGAPixelDepth_24Bit, TGAPixelDepth_32Bit);

  TColor3 = packed record
    B, G, R: Byte;
  end;

  TColor4 = packed record
    B, G, R, A: Byte;
  end;

  TPalette3 = packed array of TColor3;
  TPalette4 = packed array of TColor4;

function DumpTGA(const TGAName: AnsiString; Width, Height: Longint;
  Content: PByte; TGAPixelDepth: TTGAPixelDepth;
  TGACompression: TTGACompression = TGACompression_None): Boolean;
function DumpMappedTGA(const TGAName: AnsiString; Width, Height: Longint;
  Content: PByte; Palette: TPalette3;
  TGACompression: TTGACompression = TGACompression_None): Boolean;
function DumpIndexedTGA(const TGAName: AnsiString; Width, Height: Longint;
  Content: PByte; Palette: PByte; NumColors: Longint;
  TGAPixelDepth: TTGAPixelDepth;
  TGACompression: TTGACompression = TGACompression_None): Boolean;

implementation

uses
  SysUtils;

const
  ImageType_ColorMapped_Uncompressed = 1;
  ImageType_TrueColor_Uncompressed   = 2;
  ImageType_ColorMapped_Compressed   = 9;
  ImageType_TrueColor_Compressed     = 10;

  {$IFDEF WRITE_TGA_FOOTER}
  TGAFileFooter: packed record
    ExtensionAreaOffset     : Longword;
    DeveloperDirectoryOffset: Longword;
    Signature               : array[0..15] of Char;
    Dot                     : Char;
    ZeroTerminator          : Char;
  end =
    (ExtensionAreaOffset     : 0;
     DeveloperDirectoryOffset: 0;
     Signature               : 'TRUEVISION-XFILE';
     Dot                     : '.';
     ZeroTerminator          : #000);
  {$ENDIF}

type
  TTGAHeader = packed record
    IDLength         : Byte;
    ColorMapType     : Byte;
    ImageType        : Byte;
    FirstEntryIndex  : Word;
    ColorMapLength   : Word;
    ColorMapEntrySize: Byte;
    XOrigin          : Word;
    YOrigin          : Word;
    ImageWidth       : Word;
    ImageHeight      : Word;
    PixelDepth       : Byte;
    ImageDescriptor  : Byte;
  end;

procedure RLECompress(var InData, OutData: PByte; NumSymbols, SymSize: Longint);
var
  Len,
  Count  : Longint;
  Base,
  Last   : PByte;
  RLEMode: Boolean;
begin
  Base := InData;
  Last := PByte(Longint(InData) + (NumSymbols - 1) * SymSize);

  RLEMode := False;

  // compress while last symbol has not been passed
  while Longint(Base) <= Longint(Last) do
  begin
    InData := Base;

    if RLEMode then
    begin
      // search for a compressible sequence
      while (Longint(InData) < Longint(Last)) and CompareMem(InData,
        PByte(Longint(InData) + SymSize), SymSize) do
        Inc(InData, SymSize);
      Inc(InData, SymSize)
    end
    else
    begin
      // search for a literal run
      while (Longint(InData) < Longint(Last)) and not CompareMem(InData,
        PByte(Longint(InData) + SymSize), SymSize) do
        Inc(InData, SymSize);
      if InData = Last then
        Inc(InData, SymSize)
    end;

    // if a literal run or compressible sequence has been found, it reaches from
    // Base to InData - SymSize
    while Longint(InData) > Longint(Base) do
    begin
      // save run or sequence in blocks of no more than 128 bytes length
      Len := Longint(InData) - Longint(Base);
      if Len > $80 * SymSize then
        Len := $80 * SymSize;

      Count := Len div SymSize - 1;
      if RLEMode then
        Count := Count or $80;

      // set control byte
      Move(Count, OutData^, 1);
      Inc(OutData);

      if RLEMode then
      begin
        // copy symbol at Base
        Move(Base^, OutData^, SymSize);
        Inc(OutData, SymSize)
      end
      else
      begin
        // copy Len bytes starting from Base
        Move(Base^, OutData^, Len);
        Inc(OutData, Len)
      end;

      // go to start of next literal run or compressible sequence
      Inc(Base, Len)
    end;

    // switch between literal and compression mode
    RLEMode := not RLEMode
  end
end;

function CompressScanLines(UncompImage: PByte; var CompImage: PByte; Width,
  Height, ColorSize: Longint): Longint;
var
  i      : Longint;
  CompRun: PByte;
begin
  // reserve more than enough space: no pixel can be preceeded by more than
  // one control byte
  GetMem(CompImage, Width * Height * (ColorSize + 1));

  // compress line by line to preserve TGA scan line integrity
  CompRun := CompImage;
  for i := 1 to Height do
    RLECompress(UncompImage, CompRun, Width, ColorSize);

  // return size of compressed data
  Result := Longint(CompRun) - Longint(CompImage)
end;

function OpenTGAFile(var TGAFile: file; const TGAName: AnsiString): Boolean;
begin
  FileMode := fmOpenWrite;
  AssignFile(TGAFile, TGAName);
  {$I-}
  Rewrite(TGAFile, 1);
  {$I+}

  Result := (IOResult = 0)
end;

function DumpTGA(const TGAName: AnsiString; Width, Height: Longint;
  Content: PByte; TGAPixelDepth: TTGAPixelDepth;
  TGACompression: TTGACompression = TGACompression_None): Boolean;
var
  TGAFile  : file;
  TGAHeader: TTGAHeader;
  CompImage: PByte;
  ColorSize,
  CompSize : Longint;
begin
  Result := False;
  if not (TGAPixelDepth in [TGAPixelDepth_24Bit, TGAPixelDepth_32Bit]) or
    not (TGACompression in [TGACompression_None, TGACompression_RLE]) then
    Exit;

  if TGAPixelDepth = TGAPixelDepth_24Bit then
    ColorSize := SizeOf(TColor3)
  else
    ColorSize := SizeOf(TColor4);

  Result := OpenTGAFile(TGAFile, TGAName);
  if not Result then
    Exit;

  // fill header structures
  with TGAHeader do
  begin
    IDLength := 0;
    ColorMapType := 0;
    if TGACompression = TGACompression_RLE then
      ImageType := ImageType_TrueColor_Compressed
    else
      ImageType := ImageType_TrueColor_Uncompressed;
    FirstEntryIndex := 0;
    ColorMapLength := 0;
    ColorMapEntrySize := 0;
    XOrigin := 0;
    YOrigin := 0;
    ImageWidth := Width;
    ImageHeight := Height;
    PixelDepth := ColorSize shl 3;
    ImageDescriptor := $20
  end;

  // write header
  {$I-}
  BlockWrite(TGAFile, TGAHeader, SizeOf(TGAHeader));
  {$I+}
  Result := (IOResult = 0);
  if not Result then
  begin
    CloseFile(TGAFile);
    Exit
  end;

  // write image data and footer
  {$I-}
  if TGACompression = TGACompression_RLE then
  begin
    // if specified, compress the image data
    CompSize := CompressScanLines(Content, CompImage, Width, Height, ColorSize);
    BlockWrite(TGAFile, CompImage^, CompSize);
    FreeMem(CompImage)
  end
  else
    BlockWrite(TGAFile, Content^, Width * Height * ColorSize);
  {$IFDEF WRITE_TGA_FOOTER}
  BlockWrite(TGAFile, TGAFileFooter, SizeOf(TGAFileFooter));
  {$ENDIF}
  {$I+}
  Result := (IOResult = 0);

  CloseFile(TGAFile)
end;

function DumpMappedTGA(const TGAName: AnsiString; Width, Height: Longint;
  Content: PByte; Palette: TPalette3;
  TGACompression: TTGACompression = TGACompression_None): Boolean;
var
  TGAFile  : file;
  TGAHeader: TTGAHeader;
  CompSize : Longint;
  CompImage: PByte;
begin
  // only 8 bit pixel depth is supported in this method, thus the palette may
  // not be larger than 256 entries
  Result := (Length(Palette) <= 256);
  if not Result then
    Exit;

  Result := OpenTGAFile(TGAFile, TGAName);
  if not Result then
    Exit;

  // fill header structures
  with TGAHeader do
  begin
    IDLength := 0;
    ColorMapType := 1;
    if TGACompression = TGACompression_RLE then
      ImageType := ImageType_ColorMapped_Compressed
    else
      ImageType := ImageType_ColorMapped_Uncompressed;
    FirstEntryIndex := 0;
    ColorMapLength := Length(Palette);
    ColorMapEntrySize := SizeOf(TColor3) shl 3;
    XOrigin := 0;
    YOrigin := 0;
    ImageWidth := Width;
    ImageHeight := Height;
    PixelDepth := 8;
    ImageDescriptor := $20
  end;

  // write header and palette
  {$I-}
  BlockWrite(TGAFile, TGAHeader, SizeOf(TGAHeader));
  BlockWrite(TGAFile, Palette[0], Length(Palette) * SizeOf(TColor3));
  {$I+}
  Result := (IOResult = 0);
  if not Result then
  begin
    CloseFile(TGAFile);
    Exit
  end;

  // write image data and footer
  {$I-}
  if TGACompression = TGACompression_RLE then
  begin
    // if specified, compress the image data
    CompSize := CompressScanLines(Content, CompImage, Width, Height, 1);
    BlockWrite(TGAFile, CompImage^, CompSize);
    FreeMem(CompImage)
  end
  else
    BlockWrite(TGAFile, Content^, Width * Height);
  {$IFDEF WRITE_TGA_FOOTER}
  BlockWrite(TGAFile, TGAFileFooter, SizeOf(TGAFileFooter));
  {$ENDIF}
  {$I+}
  Result := (IOResult = 0);

  CloseFile(TGAFile)
end;

function DumpIndexedTGA(const TGAName: AnsiString; Width, Height: Longint;
  Content: PByte; Palette: PByte; NumColors: Longint;
  TGAPixelDepth: TTGAPixelDepth;
  TGACompression: TTGACompression = TGACompression_None): Boolean;
var
  i        : Longint;
  Image,
  ImageIdx : PByte;
  LargePal,
  Convert  : Boolean;
  ColorSize: Longint;
  Palette3 : TPalette3;
begin
  Result := False;
  if not (TGAPixelDepth in [TGAPixelDepth_24Bit, TGAPixelDepth_32Bit]) or
    not (TGACompression in [TGACompression_None, TGACompression_RLE]) then
    Exit;

  Result := (NumColors <= 2048);
  if not Result then
    Exit;

  LargePal := (NumColors > 256);
  {$IFDEF CONVERT_WHEN_NECESSARY}
  // images with pixel depths larger than 8 bit or palettes with an attribute
  // channel have to be expanded to true color format
  Convert := LargePal or (TGAPixelDepth = TGAPixelDepth_32Bit);
  {$ELSE}
  Convert := True;
  {$ENDIF}

  if TGAPixelDepth = TGAPixelDepth_24Bit then
    ColorSize := SizeOf(TColor3)
  else
    ColorSize := SizeOf(TColor4);

  // conversion to true color format
  if Convert then
  begin
    GetMem(Image, Width * Height * ColorSize);
    ImageIdx := Image;

    for i := 1 to Width * Height do
      // If the palette is larger than 256 entries, the image data is assumed
      // to be organized in 16 bit words, else byte data is assumed.
      if LargePal then
      begin
        if PWord(Content)^ < NumColors then
          Move(PByte(Longint(Palette) + PWord(Content)^ * ColorSize)^,
            ImageIdx^, ColorSize);
        Inc(Content, 2);
        Inc(ImageIdx, ColorSize)
      end
      else
      begin
        if Content^ < NumColors then
          Move(PByte(Longint(Palette) + Content^ * ColorSize)^, ImageIdx^,
            ColorSize);
        Inc(Content);
        Inc(ImageIdx, ColorSize)
      end;

    // save converted image
    Result := DumpTGA(TGAName, Width, Height, Image, TGAPixelDepth,
      TGACompression);
    FreeMem(Image)
  end
  else
  begin
    // Images with a pixel depth of 24 bit and a palette of no more than 256
    // colors can be directly saved as color-mapped TGAs. 
    SetLength(Palette3, NumColors);
    Move(Palette^, Palette3[0], NumColors * SizeOf(TColor3));
    Result := DumpMappedTGA(TGAName, Width, Height, Content, Palette3,
      TGACompression)
  end
end;

end.



